package at.ac.uibk.gitsearch.security.oauth2;

import at.ac.uibk.gitsearch.domain.Authority;
import at.ac.uibk.gitsearch.service.dto.AdminUserDTO;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Consumer;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.gitlab4j.api.Constants.TokenType;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.ImpersonationToken;
import org.gitlab4j.api.models.ImpersonationToken.Scope;
import org.gitlab4j.api.models.Membership;
import org.gitlab4j.api.models.User;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.oauth2.core.oidc.user.OidcUser;
import org.springframework.stereotype.Component;

@Component
public class UserDetailsFetcher {

    private static final Logger logger = LogManager.getLogger(UserDetailsFetcher.class);

    @Value("${jhipster.security.authentication.jwt.token-validity-in-seconds}")
    private long tokenValidityInSeconds;

    @Value("${application.gitlab.url}")
    private String gitlabURI;

    /**
     * just a holder class for the return of update User
     */
    public static class UserDetailsInfo {

        Optional<ImpersonationToken> userToken;
        boolean modified;

        public UserDetailsInfo(Optional<ImpersonationToken> userToken, boolean modified) {
            this.userToken = userToken;
            this.modified = modified;
        }

        public Optional<ImpersonationToken> getUserToken() {
            return userToken;
        }

        public void setUserToken(Optional<ImpersonationToken> userToken) {
            this.userToken = userToken;
        }

        public boolean isModified() {
            return modified;
        }

        public void setModified(boolean modified) {
            this.modified = modified;
        }
    }

    /**
     * fills in date from OIDC repository
     *
     * @param u
     * @param oidcUser
     * @param accessToken
     * @return
     */
    public UserDetailsInfo updateUserDetails(final AdminUserDTO u, final OidcUser oidcUser, final String gitlabAccessAdminToken) {
        boolean modified = false;
        Optional<ImpersonationToken> userToken = Optional.empty();

        try (GitLabApi gitLabApi = new GitLabApi(gitlabURI, TokenType.OAUTH2_ACCESS, gitlabAccessAdminToken)) {
            gitLabApi.enableRequestResponseLogging();

            final Set<String> authorities = new HashSet<>();
            // remove all old gitlab roles
            u.getAuthorities().stream().filter(Authority::isStandardRole).forEach(authorities::add);

            // these are only public groups :-(
            @SuppressWarnings("unchecked")
            final List<String> gitLabGroups = (List<String>) oidcUser.getAttribute("groups");
            // add gitlab groups
            if (gitLabGroups != null) {
                authorities.addAll(reduceGroups(gitLabGroups)); // only add main groups
            }

            String email = (String) oidcUser.getAttribute("email");
            User gitUser = gitLabApi.getUserApi().getUserByEmail(email);

            if (gitUser == null) {
                // try to create the user
                gitUser = createNewGitLabUserFor(oidcUser);
                String generatedPassword = generatePassword();
                gitLabApi.getUserApi().createUser(gitUser, generatedPassword, true);
                gitUser = gitLabApi.getUserApi().getUserByEmail(email);

                if (gitUser == null) {
                    // gitUser still null. Something went wrong?
                    u.setLastModifiedDate(Instant.now());
                    modified = true;
                    return new UserDetailsInfo(userToken, modified);
                }
            }
            userToken = requestGitLabUserToken(gitUser, gitLabApi);

            List<Membership> memberships = gitLabApi.getUserApi().getMemberships(gitUser.getId());
            for (Membership membership : memberships) {
                authorities.add(membership.getSourceName());
            }

            modified |= updateAttribute(gitUser.getAvatarUrl(), u.getImageUrl(), u::setImageUrl);
            modified |= updateAttribute(gitUser.getEmail(), u.getEmail(), u::setEmail);
            modified |= updateAttribute(gitUser.getName(), u.getLastName(), u::setLastName);
            if (!u.getAuthorities().equals(authorities)) {
                u.setAuthorities(authorities);
                modified = true;
            }

            if (modified) {
                u.setLastModifiedDate(Instant.now());
            }
            return new UserDetailsInfo(userToken, modified);
        } catch (GitLabApiException e) {
            logger.warn("Cannot fetch details for user {}", oidcUser.getEmail(), e);
        }

        return new UserDetailsInfo(userToken, modified);
    }

    private String generatePassword() {
        return RandomStringUtils.random(8, "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789abcdefghijklmnopqrstuvwxyz=%$-.");
    }

    private User createNewGitLabUserFor(final OidcUser oidcUser) {
        User gitUser = new User();
        String email = oidcUser.getEmail();
        gitUser.setEmail(email);
        final Object nameAttribute = oidcUser.getAttribute("name");
        String name = nameAttribute == null ? "unknown" : nameAttribute.toString();
        gitUser.setName(name);

        String userName = email.replace('@', '_');

        gitUser.setUsername(userName);
        return gitUser;
    }

    Optional<ImpersonationToken> requestGitLabUserToken(final User gitUser, final GitLabApi gitLabApi) throws GitLabApiException {
        Date expiresAt = new Date(System.currentTimeMillis() + tokenValidityInSeconds * 1000L);

        ImpersonationToken token = gitLabApi
            .getUserApi()
            .createImpersonationToken(gitUser, "TemporaryGitSearchAccessToken", expiresAt, new Scope[] { Scope.API });
        return Optional.of(token);
    }

    /**
     * unfortunatelly gitlab returns not only the main groups but also all groups in
     * the project hierarchy. That makes the list
     * very large. We need only the main groups!
     *
     * @param groups
     * @return
     */
    protected List<String> reduceGroups(List<String> groups) {
        List<String> unNormalizedGroups = new ArrayList<>(groups.size());
        unNormalizedGroups.addAll(groups);
        final Iterator<String> unNormalizedGroupsIterator = unNormalizedGroups.iterator();
        while (unNormalizedGroupsIterator.hasNext()) {
            String groupToCheck = unNormalizedGroupsIterator.next();
            for (String otherGroup : unNormalizedGroups) {
                if (
                    groupToCheck.startsWith(otherGroup) &&
                    groupToCheck.length() != otherGroup.length() && // same group, just continue
                    groupToCheck.charAt(otherGroup.length()) == '/'
                ) { // it is a real subgroup
                    unNormalizedGroupsIterator.remove();
                    break;
                }
            }
        }
        return unNormalizedGroups; // now normalized :-)
    }

    private static boolean updateAttribute(String newString, String oldString, Consumer<String> setter) {
        if (oldString == null) {
            if (newString == null) {
                return false;
            }
        } else if (oldString.equals(newString)) {
            return false;
        }
        setter.accept(newString);
        return true;
    }
}
