package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.es.model.DocumentInfo;
import at.ac.uibk.gitsearch.service.GitlabService;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.ShoppingBasketService;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketInfoDTO;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketRedirectInfoDTO;
import java.io.IOException;
import java.io.InputStream;
import org.codeability.sharing.plugins.api.ShoppingBasket;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.gitlab4j.api.GitLabApiException;
import org.mapstruct.Mapper;
import org.mapstruct.factory.Mappers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

/**
 * REST controller for managing {@link DocumentInfo}.
 */
@RestController
@RequestMapping("/api")
@Transactional
public class PluginInterfaceResource {

    private final Logger log = LoggerFactory.getLogger(PluginInterfaceResource.class);

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private ShoppingBasketService basketService;

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private SearchService searchService;

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private GitlabService gitlabService;

    /**
     * assigns a shopping basket on server, and provides a redirect link together with the shopping basket token.
     * This is called by the local frontend.
     * It has no side effects, beside storing the shopping basket for a limited time frame and return it to the requesting
     * service.
     * @param basketInfo
     * @return
     */
    @PostMapping("/pluginIF/getPluginRedirectInfos")
    public ShoppingBasketRedirectInfoDTO getRedirectInfos(@RequestBody ShoppingBasketInfoDTO basketInfo) {
        final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
        return basketService.getRedirectInfo(basketInfo, baseUrl);
    }

    /**
     * {@code SEARCH  /search/page-details} : search for the searchResults corresponding
     * to the query.
     *
     * @param query the query of the searchResult search.
     * @return the result of the search.
     */
    @PostMapping("/pluginIF/v0.1/page-details")
    public SearchResultsDTO searchPageDetails(@RequestBody org.codeability.sharing.plugins.api.search.SearchRequestDTO searchRequest)
        throws IOException {
        final Authentication gitLabAuthentication = gitlabService.getGitLabAuthentication(
            searchRequest.getUserPrincipal().getName(),
            searchRequest.getUserPrincipal().getGitLabAccessToken()
        );
        SecurityContext sc = SecurityContextHolder.getContext();
        sc.setAuthentication(gitLabAuthentication);
        log.debug("REST request to search  {}", searchRequest.getQuery());
        return searchService.searchResultPage(searchRequest.getQuery());
    }

    @Mapper
    public interface SearchResultsDTOMapper {
        SearchResultsDTOMapper INSTANCE = Mappers.getMapper(SearchResultsDTOMapper.class);

        SearchResultsDTO searchResultsToPlugin(SearchResultsDTO searchResult);
    }

    /**
     * {@code SEARCH  /search/page-details} : search for the searchResults corresponding
     *  to the query.
     *
     * @param query the query of the searchResult search.
     * @return the result of the search.
     */
    @GetMapping("/pluginIF/v0.1/basket/{basketToken}")
    public ShoppingBasket getBasket(@PathVariable("basketToken") String basketToken) {
        log.debug("REST request for basket  {}", basketToken);
        return basketService.getBasket(basketToken);
    }

    @GetMapping("/pluginIF/v0.1/basket/{basketToken}/repository/{exerciseId}")
    @SuppressWarnings("PMD.CloseResource")
    public ResponseEntity<?> getRepositoryZip(@PathVariable String basketToken, @PathVariable int exerciseId) throws IOException {
        InputStream zip;
        try {
            zip = basketService.getRepositoryZip(basketToken, exerciseId);
        } catch (GitLabApiException e) {
            return ResponseEntity.unprocessableEntity().body(e.getMessage());
        }
        InputStreamResource resource = new InputStreamResource(zip);
        return ResponseEntity
            .ok()
            .contentType(org.springframework.http.MediaType.APPLICATION_OCTET_STREAM)
            .header("filename", "Repository" + exerciseId)
            .body(resource);
    }
}
