package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.ReviewService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.dto.ReviewDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRequest;
import at.ac.uibk.gitsearch.service.dto.UserDTO;
import java.util.List;
import java.util.Optional;
import javax.transaction.Transactional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

@RestController
@RequestMapping("/api")
@Transactional
public class ReviewResource {

    private final Logger log = LoggerFactory.getLogger(ReviewResource.class);

    private final ReviewService reviewService;

    private final UserService userService;

    private final String ADMIN_ROLE = "hasAuthority(\"" + AuthoritiesConstants.ADMIN + "\")";

    private final String USER_ROLE = "hasAuthority(\"" + AuthoritiesConstants.USER + "\")";

    public ReviewResource(ReviewService reviewService, UserService userService) {
        this.reviewService = reviewService;
        this.userService = userService;
    }

    @PostMapping("/review/create")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity createReview(@RequestBody ReviewRequest reviewRequest) {
        log.debug("REST request to create a review");
        reviewService.createReview(reviewRequest);
        return ResponseEntity.ok().build();
    }

    @PutMapping("/review")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity updateReview(@RequestBody ReviewDTO review) {
        log.debug("REST request to update a review");
        reviewService.updateReview(review);
        return ResponseEntity.ok().build();
    }

    @GetMapping("/review/{id}")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity<ReviewDTO> getReview(@PathVariable Long id) {
        log.debug("REST request to get review with id: {}", id);
        return ResponseEntity.ok(reviewService.getReview(id));
    }

    //GET ALL REVIEWS FROM A USER
    @GetMapping("/review/allByUser")
    @PreAuthorize(USER_ROLE)
    public ResponseEntity<List<ReviewDTO>> getReviewsByUser() {
        return ResponseEntity.ok(reviewService.getReviewsByUser());
    }

    //UPDATE A REVIEW BY A USER
    @PutMapping("/review/updateByUser")
    @PreAuthorize(USER_ROLE)
    public ResponseEntity updateReviewRatingByUser(@RequestBody ReviewDTO review) throws UsernameNotFoundException, IllegalAccessException {
        log.debug("REST request to update a review");
        reviewService.updateReviewRatingByUser(review);
        return ResponseEntity.ok().build();
    }

    @GetMapping("/review/all")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity<List<ReviewDTO>> getAllReviews() {
        log.debug("REST request to get all reviews");
        return ResponseEntity.ok(reviewService.getAllReviews());
    }

    @DeleteMapping("/review/{id}")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity<Void> deleteReview(@PathVariable Long id) {
        log.debug("REST request to delete review: {}", id);
        reviewService.deleteReview(id);
        return ResponseEntity.ok().build();
    }

    @PostMapping("/requestReview")
    @PreAuthorize(USER_ROLE)
    public ResponseEntity requestReviewForExercise(@RequestBody ReviewRequest reviewRequest) {
        log.debug("REST request to request a review for exercise: {}", reviewRequest.getResource());
        reviewService.requestReviewForExercise(reviewRequest);
        return ResponseEntity.ok().build();
    }

    @PostMapping("/review/notifyReviewers")
    @PreAuthorize(USER_ROLE)
    public ResponseEntity notifyReviewers(@RequestBody ReviewDTO review) {
        log.debug("REST request to notify reviewers of resource: {}", review.getResource());
        reviewService.updateStatus(review.getId(), "Review improved");
        reviewService.notifyReviewers(
            review.getId(),
            "The review for " + review.getResource() + " has been updated. Please verify and update the status if applicable."
        );
        return ResponseEntity.ok().build();
    }

    @PostMapping("/review/rewardBadge")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity rewardBadge(@RequestBody ReviewDTO review) {
        log.debug("REST request to award badge for resource: {}", review.getResource());
        reviewService.rewardBadge(review.getId());
        return ResponseEntity.ok().build();
    }

    @PostMapping("/review/changeStatus")
    @PreAuthorize(USER_ROLE)
    public ResponseEntity<UserDTO> changeStatus(@RequestBody boolean status) {
        log.debug("REST request to change status of reviewer: {}", status);
        Optional<User> user = userService.getUserWithAuthorities();
        if (user.isPresent()) {
            user.get().setReviewingEnabled(status);
            userService.updateUser(user.get());
        }
        return ResponseEntity.ok().build();
    }

    //Get all reviews with statistics
    @GetMapping("/review/allWithStatistics")
    @PreAuthorize(ADMIN_ROLE)
    public ResponseEntity<List<ReviewDTO>> getAllReviewsWithStatistics() {
        log.debug("REST request to get all reviews with statistics");
        return ResponseEntity.ok(reviewService.getAllReviewsWithStatistics());
    }
}
