package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.es.model.DocumentInfo;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.StatisticsService;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.web.util.HeaderUtil;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.List;
import java.util.Optional;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

/**
 * REST controller for managing {@link DocumentInfo}.
 */
@RestController
@RequestMapping("/api")
@Transactional
public class SearchResource {

    //  @Value("${jhipster.clientApp.name}") // warning: inconsistent rename of gitSearchApp to gitSearchV2App :-(
    private static final String applicationName = "gitsearchApp";

    private static final String ENTITY_NAME = "SearchResource";

    private final Logger log = LoggerFactory.getLogger(SearchResource.class);

    private final SearchService searchService;

    private final StatisticsService statisticsService;

    private final MetaDataRepository metaDataRepository;

    public SearchResource(SearchService searchService, StatisticsService statisticsService, MetaDataRepository metaDataRepository) {
        this.searchService = searchService;
        this.statisticsService = statisticsService;
        this.metaDataRepository = metaDataRepository;
    }

    /**
     * {@code SEARCH  /search/page-details} : search for the searchResults
     * corresponding to the query.
     *
     * @param query the query of the searchResult search.
     * @return the result of the search.
     */
    @PostMapping("/search/page-details")
    public SearchResultsDTO searchPageDetails(@RequestBody org.codeability.sharing.plugins.api.search.SearchInputDTO query)
        throws IOException {
        log.debug("REST request to search  {}", query);
        return searchService.searchResultPage(query);
    }

    @PostMapping("/search/author")
    public SearchResultsDTO searchResultsByEmail(@RequestBody String email) throws IOException {
        log.debug("REST request to search  {}", email);
        return searchService.searchResultForAuthorEmail(email, 1000);
    }

    @GetMapping("/search/author")
    public SearchResultsDTO searchResultsByEmail() throws IOException {
        return searchService.searchResultsForLoggedInUser();
    }

    @GetMapping("/search/all")
    public List<SearchResultDTO> searchAll() throws IOException {
        log.debug("REST request to search for all resources");
        return searchService.getAllResources();
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param keyWordPrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/keywordsAutoComplete")
    public List<AutoCompleteEntry> getKeywordsAutoComplete(@RequestParam String keyWordPrefix, @RequestParam(defaultValue = "10") int max)
        throws IOException {
        return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix, max);
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param formatPrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/formatsAutoComplete")
    public List<AutoCompleteEntry> getFormatsAutoComplete(@RequestParam String formatPrefix, @RequestParam(defaultValue = "10") int max)
        throws IOException {
        return metaDataRepository.getFormatsAutoComplete(formatPrefix, max);
    }

    /**
     * returns all creator autocompletes
     *
     * @param creatorPrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/creatorAutoComplete")
    public List<AutoCompleteEntry> getCreatorAutoComplete(@RequestParam String creatorPrefix, @RequestParam(defaultValue = "10") int max)
        throws IOException {
        return metaDataRepository.getCreatorAutoComplete(creatorPrefix, max);
    }

    /**
     * returns all contributor autocompletes
     *
     * @param contributorPrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/contributorAutoComplete")
    public List<AutoCompleteEntry> getContributorAutoComplete(
        @RequestParam String contributorPrefix,
        @RequestParam(defaultValue = "10") int max
    ) throws IOException {
        return metaDataRepository.getContributorAutoComplete(contributorPrefix, max);
    }

    /**
     * returns all contributor and author autocompletes
     *
     * @param contributorPrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/contributorCreatorAutoComplete")
    public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix, @RequestParam(defaultValue = "10") int max)
        throws IOException {
        return metaDataRepository.getContributorCreatorAutoComplete(contributorPrefix, max);
    }

    /**
     * returns all programmingLanguage autocompletes for keyWord
     *
     * @param programmingLanguagePrefix
     * @param max the maximum number of hits returned
     * @return
     * @throws IOException
     */
    @GetMapping("/search/programmingLanguageAutoComplete")
    public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(
        @RequestParam String programmingLanguagePrefix,
        @RequestParam(defaultValue = "10") int max
    ) throws IOException {
        return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix, max);
    }

    /**
     * POST /programming-exercises/:exerciseId/export-programmingExercise : sends all repositories and details of the programming exercise as zip
     * TODO move to Exercise Resource, and export is not only for programming-exercises!
     *
     * @param exerciseId the id of the exercise to get the repos from
     * ResponseEntity with status
     * @throws IOException if something during the zip process went wrong
     */
    @PostMapping("/programming-exercises/{exerciseId}/export-programming-exercise")
    public ResponseEntity<Resource> exportProgrammingExercise(@PathVariable String exerciseId) throws IOException {
        File zipFile = null;
        try {
            zipFile = searchService.exportExercise(exerciseId);
        } catch (ParseException e) {
            log.warn("Cannot parse " + exerciseId + " exception message: {}", e);
        }
        if (zipFile == null) {
            return ResponseEntity
                .badRequest()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        ENTITY_NAME,
                        "internalServerError",
                        "There was an error on the server and the zip file could not be created."
                    )
                )
                .body(null);
        }

        /*
         * Customized FileInputStream to delete and therefore clean up the returned
         * files
         */
        class NewFileInputStream extends FileInputStream {

            File file;

            public NewFileInputStream(File file) throws FileNotFoundException {
                super(file);
                this.file = file;
            }

            @Override
            public void close() throws IOException {
                super.close();
                try {
                    Files.delete(file.toPath());
                } catch (IOException e) {
                    log.info("Cannot delete {}", file.getAbsoluteFile(), e);
                }
            }
        }
        InputStreamResource resource = new InputStreamResource(new NewFileInputStream(zipFile));

        log.debug("REST request to get Statistics for ExerciseID : {}", exerciseId);
        Optional<StatisticsDTO> statisticsDTO = statisticsService.findOneByExerciseID(exerciseId);
        if (statisticsDTO.isPresent()) {
            StatisticsDTO newStats = statisticsDTO.get();
            newStats.setDownloads(newStats.getDownloads() + 1);
            statisticsService.save(newStats);
            log.debug("REST increased number of downloads for ExerciseID : {}", exerciseId);
        }
        if (!statisticsDTO.isPresent()) {
            StatisticsDTO newStats = new StatisticsDTO();
            newStats.setDownloads(1);
            newStats.setViews(1);
            newStats.setExerciseID(exerciseId);
            statisticsService.save(newStats);
            log.debug("Created new statistics entry for exerciseID: {}", exerciseId);
        }

        return ResponseEntity
            .ok()
            .contentLength(zipFile.length())
            .contentType(MediaType.APPLICATION_OCTET_STREAM)
            .header("filename", zipFile.getName())
            .body(resource);
    }
}
